#version 430 core

// Input variables
in VertexData
{
	vec3 position_ws;	// Position in world space
	vec3 normal_ws;		// Normal in world space
	vec2 tex_coord;
	vec3 tangent_ws;	// Tangent in world space
	vec3 bitangent_ws;	// Bitangent in world space
} inData;

// Output variables
layout (location = 0) out vec4 final_color;

// Data of the camera
layout (std140, binding = 0) uniform CameraData
{
	mat4 projection;		// Projection matrix
	mat4 projection_inv;	// Inverse of the projection matrix
	mat4 view;				// View matrix
	mat4 view_inv;			// Inverse of the view matrix
	mat3 view_it;			// Inverse of the transpose of the top-left part 3x3 of the view matrix
	vec3 eye_position;		// Position of the eye in world space
};

// Data of the object
layout (binding = 0) uniform sampler2D color_tex;
layout (binding = 1) uniform sampler2D normal_tex;
uniform float height_scale;
uniform bool use_normal_mapping;

// Data of the material
layout (std140, binding = 3) uniform MaterialData
{
	// See the C++ code for the documentation to individual attributes
	vec3 ambient;
	vec3 diffuse;
	float alpha;
	vec3 specular;
	float shininess;
} material;

// Data of the light
struct PhongLight
{
	// See the C++ code for the documentation to individual attributes
	vec4 position;
	vec3 ambient;
	vec3 diffuse;
	vec3 specular;
	vec3 spot_direction;
	float spot_exponent;
	float spot_cos_cutoff;
	float atten_constant;
	float atten_linear;
	float atten_quadratic;
};
layout (std140, binding = 1) uniform PhongLightsData
{
	vec3 global_ambient_color;
	int lights_count;
	PhongLight lights[8];
};

// Evaluates the lighting of one Phong light. The implementation is at the end of the file
void EvaluatePhongLight(in PhongLight light, out vec3 amb, out vec3 dif, out vec3 spe, in vec3 normal, in vec3 position, in vec3 eye, in float shininess);

//-----------------------------------------------------------------------

void main()
{
	// Task 1: Already done, outputs white color only
	//final_color = vec4(1.0);
	//return;

	// Task 2: Comment the code above. Instead of returning a single color, return the color of the texture, still without lighting (already done)
	vec3 color_tex_val = texture(color_tex, inData.tex_coord).rgb;
	//final_color = vec4(color_tex_val, 1.0);
	//return;

	// Task 3: Comment the lines above with 'final_color' and 'return', leave only 'color_tex_val'.
	//		The code below computes the lighting (we may use it now, finally, we have the normal).

	// Compute the normal
	vec3 N;
	if (use_normal_mapping)
	{
		vec3 normal_tex_val = texture(normal_tex, inData.tex_coord).rgb * 2.0 - 1.0;
		normal_tex_val *= vec3(height_scale, height_scale, 1.0);

		// Task 6: Implement normal mapping. The normal in tangent space is already sampled from the texture,
		//				and scaled according to height_scale, so just transform it into world space and store it into 'N'
		//		Hint: Use tangent_ws, bitangent_ws, and normal_ws from inData
		vec3 T0 = normalize(inData.tangent_ws);
		vec3 B0 = normalize(inData.bitangent_ws);
		vec3 N0 = normalize(inData.normal_ws);

		N = normalize(T0 * normal_tex_val.x + B0 * normal_tex_val.y + N0 * normal_tex_val.z);
	}
	else
		N = normalize(inData.normal_ws);

	// Compute the lighting
	vec3 Eye = normalize(eye_position - inData.position_ws);

	vec3 amb = global_ambient_color;
	vec3 dif = vec3(0.0);
	vec3 spe = vec3(0.0);

	for (int i = 0; i < lights_count; i++)
	{
		vec3 a, d, s;
		EvaluatePhongLight(lights[i], a, d, s, N, inData.position_ws, Eye, material.shininess);
		amb += a;	dif += d;	spe += s;
	}

	// Compute the material
	vec3 mat_ambient = color_tex_val;
	vec3 mat_diffuse = color_tex_val;
	vec3 mat_specular = material.specular;

	// Compute the final color
	vec3 final_light = mat_ambient * amb + mat_diffuse * dif + mat_specular * spe;
	
	final_color = vec4(final_light, 1.0);
}

//-----------------------------------------------------------------------

// Evaluates the lighting of one Phong light
// light	.. [in] parameters of the light that is evaluated
// amb		.. [out] result, ambient part
// dif		.. [out] result, diffuse part
// spe		.. [out] result, specular part
// norm		.. [in] surface normal, in the world coordinates
// pos		.. [in] surface position, in the world coordinates
// eye		.. [in] direction from the surface to the eye, in the world coordinates
// shi		.. [in] shininess of the material
void EvaluatePhongLight(in PhongLight light, out vec3 amb, out vec3 dif, out vec3 spe, in vec3 norm, in vec3 pos, in vec3 eye, in float shi)
{
	vec3 L_not_normalized = light.position.xyz - pos * light.position.w;
	vec3 L = normalize(L_not_normalized);
	vec3 H = normalize(L + eye);

	// Calculate basic Phong factors
	float Iamb = 1.0;
	float Idif = max(dot(norm, L), 0.0);
	float Ispe = (Idif > 0.0) ? pow(max(dot(norm, H), 0.0), shi) : 0.0;

	// Calculate spot light factor
	if (light.spot_cos_cutoff != -1.0)
	{
		// This is a spot light
		float spot_factor;

		float spot_cos_angle = dot(-L, light.spot_direction);
		if (spot_cos_angle > light.spot_cos_cutoff)
		{
			spot_factor = pow(spot_cos_angle, light.spot_exponent);
		}
		else spot_factor = 0.0;

		Iamb *= 1.0;
		Idif *= spot_factor;
		Ispe *= spot_factor;
	}

	// Calculate attenuation
	if (light.position.w != 0.0)
	{
		// This is a point light / spot light

		float distance_from_light = length(L_not_normalized);
		float atten_factor =
			light.atten_constant +
			light.atten_linear * distance_from_light + 
			light.atten_quadratic * distance_from_light * distance_from_light;
		atten_factor = 1.0 / atten_factor;

		Iamb *= atten_factor;
		Idif *= atten_factor;
		Ispe *= atten_factor;
	}

	amb = Iamb * light.ambient;
	dif = Idif * light.diffuse;
	spe = Ispe * light.specular;
}
//#version 400
/*
layout (std140) uniform Lights
{
	vec3 light_position;		// In camera space
	vec3 light_amb;
	vec3 light_dif;
	vec3 light_spe;
};

layout (std140) uniform Material
{
	vec3 mat_amb;
	vec3 mat_dif;
	vec3 mat_spe;
	float mat_shi;
};

uniform sampler2D bumpTex;

uniform float height_scale;

in Data
{
	vec3 Position;
	vec3 Normal;
	vec3 Tangent;
	vec3 Bitangent;
	vec2 TexCoord;
} DataIn;

out vec4 outputF;

// Computes factors for Blinn-Phong lighting model. Similar to 'lit' in HLSL
vec4 lit(in float n_dot_l, in float n_dot_h, in float exponent)
{
	if (n_dot_l < 0.0)
		return vec4(1.0, 0.0, 0.0, 1.0);
	else
		if (n_dot_h < 0.0)
			return vec4(1.0, n_dot_l, 0.0, 1.0);
		else
			return vec4(1.0, n_dot_l, pow(n_dot_h, exponent), 1.0);
}
*/
//----------------------------------------------------------------
// TODO:
// 7) Comment single black color, uncomment Phong shading
// 10) Comment fixed normal, uncomment bump mapping
//----------------------------------------------------------------
/*
void main()
{
	// Black only
	// TODO 7) Comment single black color, uncomment Phong shading
	//outputF = vec4(0.0);

	// TODO 10) Comment fixed normal, uncomment bump mapping
	// Compute the derivative of the height texture. Try higher LOD or lower approximation step
	// to reduce the artefacts.
	float step_lod = 4.0;
	float height_h = 1.0 / float(textureSize(bumpTex, 0)) * step_lod;
	float sample_lod = 0.0;
	float height = textureLod(bumpTex, DataIn.TexCoord, sample_lod).x;
	float height_dx = (textureLod(bumpTex, vec2(DataIn.TexCoord.x + height_h, DataIn.TexCoord.y), sample_lod).x - height) / height_h;
	float height_dy = (textureLod(bumpTex, vec2(DataIn.TexCoord.x, DataIn.TexCoord.y + height_h), sample_lod).x - height) / height_h;

	vec3 t = DataIn.Tangent + DataIn.Normal * height_dx * height_scale;
	vec3 b = DataIn.Bitangent + DataIn.Normal * height_dy * height_scale;
	vec3 n = normalize(cross(t, b));
	
	// TODO 7) Comment single black color, uncomment Phong shading
	// TODO 10) Comment fixed normal, uncomment bump mapping
	//vec3 n = normalize(Normal);

	vec3 light_vec = normalize(light_position - DataIn.Position);
	vec3 view_vec = normalize(-DataIn.Position);
	vec3 half_vec = normalize(light_vec + view_vec);

	vec4 factors = lit(dot(n, light_vec), dot(n, half_vec), mat_shi);
	vec3 color =
		light_amb * mat_amb * factors.x +
		light_dif * mat_dif * factors.y +
		light_spe * mat_spe * factors.z;

	outputF = vec4(color, 1.0);
}*/