#version 430 core

layout(vertices = 16) out;

// Data of the camera
layout (std140, binding = 0) uniform CameraData
{
	mat4 projection;		// Projection matrix
	mat4 projection_inv;	// Inverse of the projection matrix
	mat4 view;				// View matrix
	mat4 view_inv;			// Inverse of the view matrix
	mat3 view_it;			// Inverse of the transpose of the top-left part 3x3 of the view matrix
	vec3 eye_position;		// Position of the eye in world space
};

uniform vec2 window_size;				// Size of the window in pixels
uniform float triangle_size;			// Approximative size of one triangle in pixels when auto_tessellation is true
uniform bool auto_tessellation;			// Automatic / fixed tessellation factors
uniform float tessellation_factor;		// Tessellation factor when auto_tessellation is false

// Input variables
in VertexData
{
	vec3 position_ws;	// Position in world space
	vec3 tangent_ws;	// Tangent in world space
	vec3 bitangent_ws;	// Bitangent in world space
} inData[];

// Output variables
out VertexData
{
	vec3 position_ws;	// Position in world space
	vec3 tangent_ws;	// Tangent in world space
	vec3 bitangent_ws;	// Bitangent in world space
} outData[];

// Only to pass the data between invocations, not used in the evaluation shader
out vec2 position_ss[];		// Positions in screen space (i.e. after dividing by w and after multiplying by window size

void main()
{
	// Pass the position, tangent, and bitangent from inData to outData
	outData[gl_InvocationID].position_ws = inData[gl_InvocationID].position_ws;
	outData[gl_InvocationID].tangent_ws = inData[gl_InvocationID].tangent_ws;
	outData[gl_InvocationID].bitangent_ws = inData[gl_InvocationID].bitangent_ws;

	// Task 4: Compute the position of the control point in the screen space, and store it into shared 'position_ss' array
	//		Hint: Store the result into position_ss[gl_InvocationID], don't forget to call barrier()
	vec4 position_cs = projection * view * vec4(inData[gl_InvocationID].position_ws, 1.0);		// Position in clip space
	position_ss[gl_InvocationID] = position_cs.xy / position_cs.w * window_size / 2.0;			// Position in screen space
	barrier();

	// Compute the tessellation factor
	if (gl_InvocationID == 0)
	{
		if (auto_tessellation)
		{
			// Task 4: Compute the tessellation factors
			//		Hint: function length
			vec2 p0_0 = position_ss[8]  - position_ss[12];
			vec2 p0_1 = position_ss[4]  - position_ss[8];
			vec2 p0_2 = position_ss[0]  - position_ss[4];
			vec2 p1_0 = position_ss[1]  - position_ss[0];
			vec2 p1_1 = position_ss[2]  - position_ss[1];
			vec2 p1_2 = position_ss[3]  - position_ss[2];
			vec2 p2_0 = position_ss[7]  - position_ss[3];
			vec2 p2_1 = position_ss[11] - position_ss[7];
			vec2 p2_2 = position_ss[15] - position_ss[11];
			vec2 p3_0 = position_ss[14] - position_ss[15];
			vec2 p3_1 = position_ss[13] - position_ss[14];
			vec2 p3_2 = position_ss[12] - position_ss[13];

			vec4 side_leghts = vec4(
				length(p0_0) + length(p0_1) + length(p0_2),
				length(p1_0) + length(p1_1) + length(p1_2),
				length(p2_0) + length(p2_1) + length(p2_2),
				length(p3_0) + length(p3_1) + length(p3_2));

			// ..., now the number of splits, clamp it to some reasonable numbers
			vec4 out_factors = side_leghts / triangle_size;
			out_factors = clamp(out_factors, 1.0f, 64.0f);		// Clamp the result, make sure there is at least one split

			// Set the tessellation levels
			gl_TessLevelInner[0] = (out_factors[1] + out_factors[3]) / 2.0;
			gl_TessLevelInner[1] = (out_factors[0] + out_factors[2]) / 2.0;
			gl_TessLevelOuter[0] = out_factors[0];
			gl_TessLevelOuter[1] = out_factors[1];
			gl_TessLevelOuter[2] = out_factors[2];
			gl_TessLevelOuter[3] = out_factors[3];
		}
		else
		{
			// Set all tessellation factors to 'tessellation_factor'.
			gl_TessLevelInner[0] = tessellation_factor;
			gl_TessLevelInner[1] = tessellation_factor;
			gl_TessLevelOuter[0] = tessellation_factor;
			gl_TessLevelOuter[1] = tessellation_factor;
			gl_TessLevelOuter[2] = tessellation_factor;
			gl_TessLevelOuter[3] = tessellation_factor;
		}
	}
}